use std::{error::Error, fs::File, io::Write, path::Path};
use clap::Parser;
use crate::reskit::{tileset, soundtrack::{formats::dmf::DmfModule, engines::echo::engine::{EchoFormat, EchoArtifact}}, utility::print_good};
use super::settings::{Args, Tools, TileOutputFormat, TileOrder};

pub fn run_command() -> Result<(), Box<dyn Error>> {
    let cli = Args::parse();

    match cli.tool {
        Tools::Tileset { input_file, output_file, format, tile_order } =>
            tileset::generate(
                &input_file,
                &output_file,
                match format {
                    TileOutputFormat::Bin => "bin",
                    TileOutputFormat::Inc => "inc"
                },
                match tile_order {
                    TileOrder::Tile => "tile",
                    TileOrder::Sprite => "sprite"
                }
            ),
        Tools::Soundtrack { input_files, output_directory, input_format: _, output_format: _, source_file_format: _, artifact_output_directory } => {
            if input_files.is_empty() {
                return Err( "no input files (see `reskit soundtrack --help` for more info)" )?;
            }

            let output_directory = if output_directory.ends_with( "/" ) { output_directory.to_string() } else { format!( "{}/", output_directory ) };
            let artifact_output_directory = if artifact_output_directory.ends_with( "/" ) { artifact_output_directory.to_string() } else { format!( "{}/", artifact_output_directory ) };

            let modules: Vec<DmfModule> = input_files.iter().map( | filename | Ok( DmfModule::from_file( &filename )? ) ).collect::<Result<Vec<DmfModule>, Box<dyn Error>>>()?;

            let mut combined_asset_list = DmfModule::get_combined_assets_list( artifact_output_directory.clone(), &modules )?;
            if combined_asset_list.instruments.len() > 255 {
                return Err( "combined instrument set exceeds 255 instruments; try converting fewer modules or modules with fewer instruments" )?;
            }
            if combined_asset_list.samples.len() > 255 {
                return Err( "combined sample set exceeds 255 samples; try converting fewer modules or modules with fewer samples" )?;
            }

            // Output the ESF sequences
            for i in 0..modules.len() {
                let module = &modules[ i ];

                let path = Path::new( &module.path );
                let original_filename_stem = path.file_stem().ok_or( "internal error: unable to parse filename" )?.to_string_lossy();
                let output_file = format!( "{}{}.esf", output_directory, original_filename_stem );

                combined_asset_list.sequences.push( output_file.clone() );

                println!( "Creating ESF output file {}", output_file );
                let mut file = File::create( &output_file )?;
                file.write_all( &module.get_esf( &combined_asset_list )? )?;

                print_good( &format!( "successfully compiled soundtrack for {}", output_file ) );
            }

            // Output the shared artifacts used by all ESF sequences
            println!( "Writing sequence artifacts..." );
            let shared_artifacts = combined_asset_list.to_bytes()?;
            println!( "Writing sequence artifact {}", format!( "{}music.asm", artifact_output_directory ) );
            let mut file = File::create( format!( "{}music.asm", artifact_output_directory ) )?;
            file.write_all( &shared_artifacts )?;
        }
    };

    Ok( () )
}